# Praat script evoContrast/simulate.praat
# Paul Boersma, January 26, 2008

form Simulate the evolution of contrast
	real Minimum_frequency_(Erb) 20.0
	real Maximum_frequency_(Erb) 36.0
	positive Frequency_step_(Erb) 0.1
	natural Number_of_categories 2
	real Mean1 25.5
	real Mean2 30.5
	real Mean3 200
	real Mean4 200
	positive Initial_production_noise 0.8
	positive Transmission_noise 0.8
	positive Neighbourhood 0.5
	natural Number_of_trials 2000000
	natural Number_of_generations 10
	choice Decision_strategy 1
		option Optimality Theory
		option Harmonic Grammar
		option Maximum Entropy
endform
; 3067, 4733, 7533:   24.47, 27.95, 31.37
evaluationNoise = 2.0
plasticity = 0.01
numberOfDivisions = round ((maximum_frequency - minimum_frequency) / frequency_step) + 1
Times
12
#
# Create the initial language.
#
initialLanguage = Create Table with column names... language numberOfDivisions frequency
Formula... frequency minimum_frequency + (row - 1) * frequency_step
for category to number_of_categories
	Append column... p'category'
	factor = -0.5 / initial_production_noise ^ 2
	Formula... p'category' exp (factor * (self ["frequency"] - mean'category') ^ 2)   ; distribution of frequency values for the category
endfor
for generation to number_of_generations
	#
	# Create a baby learner.
	#
	childGrammar = Create Matrix... grammar'generation' minimum_frequency maximum_frequency numberOfDivisions frequency_step minimum_frequency
	... 1 number_of_categories+1 number_of_categories+1 1 1 100
	disharmonies = To Sound (slice)... 1
	#
	# Feed data.
	#
	for trial to number_of_trials
		#
		# The parent produces a phonological category. Each has equal chance.
		#
		intendedCategory = randomInteger (1, number_of_categories)
		#
		# Production noise: the parent chooses a phonetic frequency.
		#
		if generation = 1
			select initialLanguage
			producedRow = Draw row from distribution... p'intendedCategory'
			producedFrequency = Get value... producedRow frequency
		else
			select disharmonies
			if decision_strategy$ = "Optimality Theory"
				Formula... max (
				... randomGauss (Object_'parentGrammar' [intendedCategory, col], evaluationNoise),
				... randomGauss (Object_'parentGrammar' [number_of_categories + 1, col], evaluationNoise))
				producedFrequency = Get time of minimum... 0 0 None
			elsif decision_strategy$ = "Harmonic Grammar"
				Formula...
				... randomGauss (Object_'parentGrammar' [intendedCategory, col], evaluationNoise) +
				... randomGauss (Object_'parentGrammar' [number_of_categories + 1, col], evaluationNoise)
				producedFrequency = Get time of minimum... 0 0 None
			elsif decision_strategy$ = "Maximum Entropy"
				# Harmony:
				Formula... Object_'parentGrammar' [intendedCategory, col] + Object_'parentGrammar' [number_of_categories + 1, col]
				# Relative probability:
				Formula... self [col - 1] + exp (- self)
				# Choose:
				sum = Get maximum... 0 0 None
				pick = randomUniform (0, sum)
				Reverse
				Formula... ( self [col + 1] > pick ) <> ( self [col] > pick )
				Reverse
				maximum = Get maximum... 0 0 None
				assert maximum = 1.0   ; 'sum' 'pick'
				producedFrequency = Get time of maximum... 0 0 None
			else
				exit Unknown decision strategy "'decision_strategy$'".
			endif
			; the optimal candidate
		endif
		#
		# Transmission.
		#
		receivedFrequency = randomGauss (producedFrequency, transmission_noise)
		#
		# Determine the winning candidate in perception.
		#
		ivalue = max (min (round ((receivedFrequency - minimum_frequency) / frequency_step) + 1, numberOfDivisions), 1)   ; determine the two violated constraints
		if decision_strategy$ = "Maximum Entropy"
			assert number_of_categories = 2
			p1 = exp (- Object_'childGrammar' [1, ivalue])
			p2 = exp (- Object_'childGrammar' [2, ivalue])
			pick = randomUniform (0, p1 + p2)
			perceivedCategory = if pick <= p1 then 1 else 2 fi
		else
			if number_of_categories = 1
				perceivedCategory = imin (
				... randomGauss (Object_'childGrammar' [1, ivalue], evaluationNoise))   ; the optimal candidate
			elsif number_of_categories = 2
				perceivedCategory = imin (
				... randomGauss (Object_'childGrammar' [1, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [2, ivalue], evaluationNoise))   ; the optimal candidate
			elsif number_of_categories = 3
				perceivedCategory = imin (
				... randomGauss (Object_'childGrammar' [1, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [2, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [3, ivalue], evaluationNoise))   ; the optimal candidate
			elsif number_of_categories = 4
				perceivedCategory = imin (
				... randomGauss (Object_'childGrammar' [1, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [2, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [3, ivalue], evaluationNoise),
				... randomGauss (Object_'childGrammar' [4, ivalue], evaluationNoise))   ; the optimal candidate
			endif
		endif
		#
		# Learn perception.
		#
		if intendedCategory <> perceivedCategory
			select childGrammar
			factor = -0.5 / neighbourhood ^ 2
			Formula... if row = intendedCategory then self - plasticity * exp (factor * (x - receivedFrequency) ^ 2)
			... else if row = perceivedCategory then self + plasticity * exp (factor * (x - receivedFrequency) ^ 2)
			... else self fi fi   ; gradual learning algorithm weighted by degree of violation (as in Boersma 1997)
		endif
	endfor
	#
	# Create production grammar.
	#
	select childGrammar
	Formula... if row = nrow then 91 + 13 * ((x - (minimum_frequency + maximum_frequency) / 2) / ((maximum_frequency - minimum_frequency) / 2)) ^ 2 else self fi   ; articulatory constraints
	execute drawGrammar.praat 1
	#
	# Let time go by.
	#
	parentGrammar = childGrammar
	select disharmonies
	Remove
endfor
